'''

Evolution is LIT 

    - Darwin, one can only presume

'''


import numpy as np

from freelunch.adaptable import adaptable_method
from freelunch.util import BadObjectiveFunctionScores


# %% Base class


class genetic_operation(adaptable_method):
    '''
    Base class for anything that happens while listening to Barry white...
    '''
    type = None
    name = 'genetic operation'
    hypers = {}

    def __init__(self, hypers=hypers):
        super().__init__()
        self.hypers = hypers

    def op(self, *parents, **hypers):
        raise NotImplementedError 


# %% Search operations
    

class rand_1(genetic_operation):
    type = 'mutation'
    name = 'rand/1'
    n_parents = 0
    hypers = {'F': 0.5}

    def op(self, *parents, pop=None, F=hypers['F']):
        idxs = np.random.randint(0, len(pop), 3)
        a, b, c = pop[idxs]
        # Assumes numpy array vectorisation
        return a.dna+ F * (b.dna - c.dna)


class rand_2(genetic_operation):
    type = 'mutation'
    name = 'rand/2'
    n_parents = 0
    hypers = {'F': 0.5}

    def op(self, *parents, pop=None, F=hypers['F']):
        idxs = np.random.randint(0, len(pop), 5)
        a, b, c, d, e = pop[idxs]
        return a.dna + F * (b.dna - c.dna) + F * (d.dna - e.dna)


class best_1(genetic_operation):
    type = 'mutation'
    name = 'best/1'
    n_parents = 0
    hypers = {'F': 0.5}

    def op(self, *parents, pop=None, F=hypers['F']):
        idxs = np.random.randint(0, len(pop), 2)
        a, b = pop[idxs]
        best = min(pop, key=lambda x: x.fitness)
        return best.dna + F * (a.dna - b.dna)


class best_2(genetic_operation):
    type = 'mutation'
    name = 'best/2'
    n_parents = 0
    hypers = {'F': 0.5}

    def op(self, *parents, pop=None, F=hypers['F']):
        idxs = np.random.randint(0, len(pop), 4)
        a, b, c, d = pop[idxs]
        best = min(pop, key=lambda x: x.fitness)
        return best.dna + F * (a.dna - b.dna) + F * (c.dna - d.dna)


class current_1(genetic_operation):
    type = 'mutation'
    name = 'current/1'
    n_parents = 1
    hypers = {'F': 0.5}

    def op(self, *parents, pop=None, F=hypers['F']):
        x = parents[0]
        idxs = np.random.randint(0, len(pop), 3)
        a, b, c = pop[idxs] 
        best = min(pop, key=lambda x: x.fitness)
        return x.dna + F * (best.dna - a.dna) + F * (b.dna - c.dna)



# %% Crossover operations


class binary_crossover(genetic_operation):
    type = 'crossover'
    name='binary crossover'
    hypers={'Cr':0.2}

    def op(self, parent1, parent2, Cr=hypers['Cr']):
        out = np.empty_like(parent1)
        for a, b, i in zip(parent1, parent2, range(len(parent1))):
            if np.random.uniform(0, 1) < Cr:
                out[i] = a
            else:
                out[i] = b
        jrand = np.random.randint(0, len(out)) #Ensure at least one difference
        out[jrand] = parent2[jrand]
        return out


# %% Selection operations


class binary_tournament(genetic_operation):
    '''
    2 - tournament selection
    '''
    type = 'selection'
    name = 'binary tournament'
    hypers={}

    def op(self, olds, news):
        out = np.empty_like(olds, dtype=object)
        for old, new, i in zip(olds, news, range(len(out))):
            if new < old:
                out[i] = new
                new.on_win()
            else:
                out[i] = old
        return out



'''
Benchmarks for testing / comparisons
'''

import numpy as np
from freelunch.util import ZeroLengthSolutionError

# %%

class benchmark:
    '''
    base class for benchmarking functions
    '''
    default_bounds = lambda n:None
    rtm_optimum = lambda n:None
    tol=None

    def __init__(self, n=2):
        self.n = n
        self.bounds = self.default_bounds() 
        self.optimum = self.rtn_optimum()

    def __call__(self, dna):
        if len(dna) == 0:
            raise ZeroLengthSolutionError('An empty trial solution was passed')
        return self.obj(dna)

# %% some misc (v0.x) benchmarks

class ackley(benchmark):
    '''
    ackely function in n dimensions
    '''

    default_bounds = lambda self:np.array([[-10, 10]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = 0
    tol = 10**-2

    a,b,c = 20, 0.2, 2*np.pi
    def obj(self, dna):
        t1 = -self.a * np.exp(-self.b * (1/len(dna)) * np.sum(dna**2))
        t2 = - np.exp(1/len(dna) * np.sum(np.cos(self.c * dna))) 
        t3 = self.a + np.exp(1)
        return t1 + t2 + t3


class exponential(benchmark):
    '''
    exponential function in n dimensions
    '''
    
    default_bounds = lambda self:np.array([[-10, 10]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = -1
    tol = 10**-3

    a = -0.5
    def obj(self, dna):
        t1 = - np.exp(self.a * np.sum(dna**2))
        return t1


class happycat(benchmark):
    '''
    happycat function in n dimensions
    '''
    
    default_bounds = lambda self:np.array([[-2, 2]]*self.n)
    rtn_optimum = lambda self:np.array([-1]*self.n)
    f0 = 0
    tol = None

    a = 1/8
    def obj(self, dna):
        norm = np.sum(dna**2)
        t1 = ((norm - self.n)**2)**(self.a)
        t2 = (1/self.n)*(0.5*norm + np.sum(dna)) + 0.5
        return t1 + t2


class periodic(benchmark):
    '''
    periodic function in n dimensions
    '''
    
    default_bounds = lambda self:np.array([[-10, 10]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = 0.9
    tol = None

    def obj(self, dna):
        t1 = np.sum(np.sin(dna)**2) +1
        t2 = - 0.1 * np.exp(-np.sum(dna**2))
        return t1 + t2



# %% https://robertmarks.org/Classes/ENGR5358/Papers/functions.pdf

class DeJong(benchmark):
    '''
    DeJong's 1st function in n dimensions
    '''
    default_bounds = lambda self:np.array([[-5.12, 5.12]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = 0

    def obj(self, dna):
        return np.sum(dna**2)

class HyperElipsoid(benchmark):
    '''
    HyperElipsoid function in n dimensions
    '''
    default_bounds = lambda self:np.array([[-5.12, 5.12]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = 0

    def obj(self, dna):
        return np.sum(np.arange(1,self.n+1)*dna**2)

class RotatedHyperElipsoid(benchmark):
    '''
    RotatedHyperElipsoid function in n dimensions
    '''
    default_bounds = lambda self:np.array([[-65.536, 65.536]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = 0

    def obj(self, dna):
        out = 0
        for i in range(self.n):
            out += np.sum(dna[:i+1]**2)
        return out

class Rosenbrock(benchmark):
    '''
    Rosenbrock's function in n dimensions (banana function)
    '''
    default_bounds = lambda self:np.array([[-2.048, 2.048]]*self.n)
    rtn_optimum = lambda self:np.array([1]*self.n)
    f0 = 0

    def obj(self, dna):
        return np.sum(100*(dna[1:] - dna[:-1]**2)**2 + (1-dna[:-1])**2)
        

class Ragstrin(benchmark):
    '''
    Ragstrin's function in n dimensions 
    '''
    default_bounds = lambda self:np.array([[-5.12, 5.12]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = 0
    
    def obj(self, dna):
        return 10 * self.n + np.sum(dna**2 - 10*np.cos(2* np.pi*dna))
        
class Schwefel(benchmark):
    '''
    Schwefel's function in n dimensions

    (divided through by dimension for constant f0)
    '''
    default_bounds = lambda self:np.array([[-500, 500]]*self.n)
    rtn_optimum = lambda self:np.array([420.9687]*self.n)
    f0 = 0
    
    def obj(self, dna):
        return 418.9828872721625-np.sum(dna*np.sin(np.sqrt(np.abs(dna))))/self.n
        
class Griewangk(benchmark):
    '''
    Griewangk's function in n dimensions

    '''
    default_bounds = lambda self:np.array([[-600, 600]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = 0
    
    def obj(self, dna):
        return (1/4000)* np.sum(dna**2) - np.prod(np.cos(dna/np.sqrt(np.arange(self.n)+1))) + 1 

class PowerSum(benchmark):
    '''
    Powersum function in n dimensions

    '''
    default_bounds = lambda self:np.array([[-1, 1]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = 0
    
    def obj(self, dna):
        out = 0
        for i,x in enumerate(dna):
            out+=np.abs(x)**(i+1) 
        return out

class Ackley(benchmark):
    '''
    Ackely function in n dimensions
    '''

    default_bounds = lambda self:np.array([[-32.768, 32.768]]*self.n)
    rtn_optimum = lambda self:np.array([0]*self.n)
    f0 = 0

    a,b,c = 20, 0.2, 2*np.pi
    def obj(self, dna):
        t1 = -self.a * np.exp(-self.b * (1/len(dna)) * np.sum(dna**2))
        t2 = - np.exp(1/len(dna) * np.sum(np.cos(self.c * dna))) 
        t3 = self.a + np.exp(1)
        return t1 + t2 + t3   


MOLGA_TEST_SUITE = [
    DeJong,
    HyperElipsoid,
    RotatedHyperElipsoid,
    Rosenbrock,
    Ragstrin,
    Schwefel,
    Griewangk,
    PowerSum,
    Ackley
]


"""
Main module definitions in here
"""
import warnings

import numpy as np

from freelunch import tech, zoo, util
from freelunch.base import continuous_space_optimiser
from freelunch.adaptable import normally_varying_parameter


# %% Optimisation classes


class DE(continuous_space_optimiser):
    '''
    Differential evolution
    '''
    name = 'Differential Evolution'
    tags = ['continuous domain', 'population based', 'evolutionary']
    hyper_definitions = {
        'N': 'Population size (int)',
        'G': 'Number of generations (int)',
        'F': 'Mutation parameter (float in [0,1])',
        'Cr': 'Crossover probability (float in [0,1])',
        'mutation': 'Mutation method (str, see docs for options)',
        'crossover': 'Crossover method (str, see docs for options)',
        'selection': 'Selection method (str, see docs for options)'
    }
    hyper_defaults = {
        'N': 100,
        'G': 100,
        'F': 0.5,
        'Cr': 0.2,
        'mutation': 'rand_1',
        'crossover': 'binary_crossover',
        'selection': 'binary_tournament'
    }

    def run(self):
        self.curve_callback()
        # Parse hyperparameter
        mutator = self.parse_hyper(self.hypers['mutation'])
        breeder = self.parse_hyper(self.hypers['crossover'])
        # and the crowd says bo...                                           <-- Funniest joke in repo
        selector = self.parse_hyper(self.hypers['selection'])
        # initial pop
        pop = self.initialiser(self.bounds, self.hypers['N'])
        tech.compute_obj(pop, self.obj)
        # main loop
        for gen in range(self.hypers['G']):
            # generate trial population
            trial_pop = np.empty_like(pop, dtype=object)
            for i, sol in enumerate(pop):
                # Mutation operation
                trial = zoo.animal()
                trial.dna = mutator(sol, pop=pop, F=self.hypers['F'])
                # Crossover operation
                trial.dna = breeder(sol.dna, trial.dna, Cr=self.hypers['Cr'])
                trial_pop[i] = trial
            # Apply bounds
            self.apply_bounds(trial_pop)
            # Selection operation
            tech.compute_obj(trial_pop, self.obj)
            pop = selector(pop, trial_pop)
            scores = [p.fitness for p in pop]
            self.curve_callback(min(scores), np.mean(scores))
        return pop


class SADE(continuous_space_optimiser):
    '''
    Self-Adapting Differential evolution
    '''
    name = 'Self-Adapting Differential Evolution'
    tags = ['continuous domain', 'population based', 'evolutionary', 'adaptive']
    hyper_definitions = {
        'N': 'Population size (int)',
        'G': 'Number of generations (int)',
        'F_u': 'Mutation parameter initial mean (float in [0,2])',
        'F_sig': 'Mutation parameter initial standard deviation (float in [0,1])',
        'Cr_u': 'Crossover probability initial mean (float in [0,1])',
        'Cr_sig': 'Crossover probability initial standard deviation (float in [0,1])',
        'Lp': 'Learning period',
        'mutation': 'Mutation method (str, see docs for options)',
        'crossover': 'Crossover method (str, see docs for options)',
        'selection': 'Selection method (str, see docs for options)'
    }
    hyper_defaults = {
        'N': 100,
        'G': 100,
        'F_u': 0.5,
        'F_sig': 0.2,
        'Cr_u': 0.2,
        'Cr_sig': 0.1,
        'Lp': 10,
        'mutation': ['rand_1', 'rand_2', 'best_1', 'best_2', 'current_1'],
        'crossover': 'binary_crossover',
        'selection': 'binary_tournament'
    }

    def run(self):
        self.curve_callback()
        # initial params and operations
        mutators = self.parse_hyper(self.hypers['mutation'])
        breeder = self.parse_hyper(self.hypers['crossover'])
        selector = self.parse_hyper(self.hypers['selection'])
        F = normally_varying_parameter(
            self.hypers['F_u'], self.hypers['F_sig'])
        Cr = normally_varying_parameter(
            self.hypers['Cr_u'], self.hypers['Cr_sig'])
        # initial pop
        pop = self.initialiser(self.bounds, self.hypers['N'])
        tech.compute_obj(pop, self.obj)
        # main loop
        for gen in range(self.hypers['G']):
            # adaptable parameters/ methods update
            if gen > 0 and gen % self.hypers['Lp'] == 0:
                mutators.update_strategy_ps()
                F.update()
                Cr.update()
            # trial pop
            trial_pop = np.empty_like(pop, dtype=object)
            for i, sol in enumerate(pop):
                # mutation
                new = zoo.animal()
                mutator = mutators.select_strategy()
                new.dna = mutator(sol, pop=pop, F=F())
                # crossover
                new.dna = breeder(sol.dna, new.dna, Cr())
                # Record methods used to produce new individual
                new.tech.extend([mutator, F.now(), Cr.now()])
                trial_pop[i] = new
            # Apply bounds
            self.apply_bounds(trial_pop)
            # Selection operation
            tech.compute_obj(trial_pop, self.obj)
            pop = selector(pop, trial_pop)
            scores = [p.fitness for p in pop]
            self.curve_callback(min(scores), np.mean(scores))
        return pop


class SA(continuous_space_optimiser):
    '''
    Simulated Annealing 
    '''
    name = 'Simulated Annealing'
    tags = ['Continuous domain', 'Annealing']
    hyper_definitions = {
        'N': 'number of (independent) runs',
        'K': 'number of timesteps (int)',
        'T0': 'Max Temperature (float)',
    }
    hyper_defaults = {  # Super simple and prescriptive for now.
        'N': 100,
        'K': 1500,
        'T0': 50,
    }

    def P(self, e_old, e_new, T):  # i.e. M-H, Kirkpatrick et al.
        if e_new is None:
            e_new = e_old+1  # Clumsy but works
        if e_new < e_old:
            return 1
        else:
            return np.exp(-(e_new - e_old) / T)

    def T(self, k):  # logistic hardcoded for now
        return self.hypers['T0']/np.log(k)

    def neighbour(self, old):  # gaussian perturbation with sticky bounds
        new = zoo.animal()
        new.dna = np.empty_like(old.dna)
        idxs = np.random.choice(len(old.dna), 1, replace=False)
        for i, u, bound in zip(idxs, old.dna, self.bounds):
            sig = (bound[1] - bound[0]) / 10
            new.dna[i] = np.random.normal(u, sig)
        return new

    def run(self):
        # initialise
        old = self.initialiser(self.bounds, self.hypers['N'])
        tech.compute_obj(old, self.obj)
        best = min(old, key=lambda x: x.fitness)
        # main loop
        for k in range(1, self.hypers['K']):
            # generate temperature
            T = self.T(k+1)
            new = np.empty_like(old)
            for i, o in enumerate(old):
                # generate neighbour
                new[i] = self.neighbour(o)
            self.apply_bounds(new)
            tech.compute_obj(new, self.obj)
            # selection with probability P
            for i, o, n in zip(range(self.hypers['N']), old, new):
                if self.P(o.fitness, n.fitness, T) >= np.random.uniform(0, 1):
                    old[i] = new[i]
                if n < best:
                    best = n
        # This is subtle, best is not neccesarily in new...
        final_pop = sorted(old)
        final_pop[-1] = best  # replace worst of new pop with best sol
        return final_pop


class PSO(continuous_space_optimiser):
    '''
    Base Class for Particle Swarm Optimisations
    '''
    name = 'Particle Swarm Optimisation'
    tags = ['Continuous domain', 'Particle swarm']
    hyper_definitions = {
        'N':'Population size (int)',
        'G':'Number of generations (int)',
        'I':'Inertia coefficients (np.array, I.shape=(2,))',
        'A':'Acceleration (np.array, I.shape=(2,))'
    }
    hyper_defaults = {
        'N':100,
        'G':200,
        'I':np.array([0.1, 0.9]),
        'A':np.array([0.1, 0.1])
    }

    def init_pop(self,N):
        # Function which initialises the swarm
        pop = np.empty((N,), dtype=object)
        for i in range(N):
            pop[i] = zoo.particle(np.array([np.random.uniform(a,b) for a, b in self.bounds]))
            pop[i].vel = np.squeeze((2*np.random.rand(self.bounds.shape[0],1)-1)*np.diff(self.bounds))
        return pop

    def move_swarm(self, pop, gen):
        # Basic particle swarm move
        inertia = self.hypers['I'][1]-(self.hypers['I'][1]-self.hypers['I'][0])*gen/self.hypers['G']
        # This loop is slow, should vectorise at some point
        for p in pop:
            p.vel = inertia*p.vel + \
                self.hypers['A'][0]*np.random.rand()*(p.best_pos-p.pos) + \
                self.hypers['A'][1]*np.random.rand()*(self.g_best.pos-p.pos)
            p.pos = p.pos + p.vel
            tech.sticky_bounds(p, self.bounds)
        return pop

    def test_pop(self, pop):
        # Test all population and update bests
        tech.compute_obj(pop, self.obj)
        return pop

    def best_particle(self, pop):
        # Particles particles on the wall, who's the bestest of them all
        best = zoo.particle()
        loc = np.squeeze(np.where(pop==min(pop,key=lambda x: x.best)))
        best.fitness = pop[loc].best
        best.pos = pop[loc].best_pos
        return best

    # Generic PSO run routine
    def run(self):
        self.curve_callback()
        # Initialise the swarm
        pop = self.init_pop(self.hypers['N'])
        # Test Initial Population
        self.test_pop(pop)
        self.g_best = self.best_particle(pop)
        # Main loop
        for gen in range(self.hypers['G']):
            # Propagate the swarm
            pop = self.move_swarm(pop,gen)
            # Test new swarm locations
            self.test_pop(pop)
            # Particle class updates best previous position
            # Update global best
            self.g_best = self.best_particle(pop)
            self.curve_callback(self.g_best.best, np.mean([p.best for p in pop]))
        return pop


class QPSO(PSO):
    '''
    Quantum Particle Swarm Optimisations
    '''
    name = 'Quantum Particle Swarm Optimisation'
    tags = ['Continuous domain', 'Particle swarm']
    hyper_definitions = {
        'N': 'Population size (int)',
        'G': 'Number of generations (int)',
        'alpha': 'Contraction Expansion Coefficient (np.array, I.shape=(2,))',
    }
    hyper_defaults = {
        'N': 100,
        'G': 200,
        'alpha': np.array([1.0, 0.5]),
    }

    def mean_best_pos(self, pop):
        return np.mean([p.best_pos for p in pop], axis=0)

    def move_swarm(self, pop, gen):
        C = self.mean_best_pos(pop)
        D = len(C)
        g_best = self.best_particle(pop)
        alpha = tech.lin_reduce(self.hypers['alpha'], gen, self.hypers['G'])
        for p in pop:
            phi = np.random.random_sample(D)
            u = np.random.random_sample(D)
            pp = phi*p.best_pos + (1-phi)*g_best.pos
            # In algorithm rand(0,1) > 0.5 is 50/50 chance
            # Replace here with np.sign(np.random.normal(size=D))
            # Also 50/50 but nicer for the algorithm
            p.pos = pp + \
                np.sign(np.random.normal(size=D)) *\
                alpha*np.abs(C - p.pos)*np.log(1/u)
        self.apply_bounds(pop)
        return pop


class PAO(continuous_space_optimiser):
    '''
    Particle Attractor Optimisation (PAO) for now.

    Essentially a state-space implementation of a second order SDE with a number of weighted attractors
    '''
    name = 'Particle Attractor Optimisation'
    tags = ['Continuous domain', 'Particle swarm']
    hyper_definitions = {
        'N': 'Population size (int)',
        'G': 'Number of generations (int)',
        'm': 'Inertia coefficient (i.e mass)',
        'c': 'Damping (ie prop. to stiffness matrix)',
        'k': 'Stiffness parmaeters (i.e lambda1, lambda2 etc.)',
        'q': 'Randomness factor (space dust)',
        'dt': 'Timestep size'
    }
    hyper_defaults = {
        'N': 100,
        'G': 100,
        'm': 1,
        'z': 0.2,
        'k': np.array([1, 1]),
        'q': [0, 0],
        'dt': 1,
        'other attractors': []
    }
    Attractors = {
        'average particle': lambda X, l, g: np.ones_like(X) * np.mean(X, axis=0),
        'average local best': lambda X, l, g: np.ones_like(X) * np.mean(l, axis=0),
    }

    # Override this function to implement custom attractors
    def compute_attractors(self, X, local_bests, global_best):
        a1 = local_bests
        a2 = np.ones_like(X) * global_best
        A = [a1, a2]
        for a in self.hypers['other attractors']:
            A.append(self.Attractors[a](X, local_bests, global_best))
        return A

    def update_attractors(self, X, prev_scores, local_bests):
        scores = np.array([self.obj(a) for a in X])
        idx = scores < prev_scores
        scores[~idx] = prev_scores[~idx]
        local_bests[idx] = X[idx]
        global_best = local_bests[np.argmin(scores)]
        # attractors
        attractors = self.compute_attractors(X, local_bests, global_best)
        attractors = np.stack(attractors, axis=2).transpose(2, 0, 1)
        # offset
        offset = np.dot(
            attractors.T, self.hypers['k']).T / np.sum(self.hypers['k'])
        return attractors, scores, offset

    def run(self):
        self.curve_callback()

        # init params
        D = len(self.bounds)
        N = self.hypers['N']
        k = np.array(self.hypers['k'])
        kp = np.sum(k)
        bounds = tech.bounds_as_mat(self.bounds)

        # init dynamics (using fraction decomposition trick)
        A = np.array([[0, 1], [-kp/self.hypers['m'], -(self.hypers['z']
                                                       * (2*np.sqrt(kp*self.hypers['m'])))/self.hypers['m']]])
        L = np.array([[0], [1]])
        Phi = util.expm(
            np.block([[A, L@L.T], [np.zeros_like(A), -A.T]])*self.hypers['dt'])
        Adt = Phi[:2, :2]
        Sig_L = np.linalg.cholesky(Phi[:2, 2:]@Adt.T)

        # init population
        X = np.random.uniform(*bounds.T, size=(N, D))
        V = np.random.normal(0, 1, size=(N, D))
        Xprime = np.stack((X, V), axis=2) # preallocating X only

        # test and update centers
        attractors, scores, offset = self.update_attractors(
            X, np.ones(N)*np.inf, X)

        # main loop
        for g in range(self.hypers['G']):
            # compute scaling factor of noise for each state
            qq = tech.lin_reduce(self.hypers['q'], g, self.hypers['G'])
            q = np.exp(qq)*(np.abs(attractors[0] - attractors[1]))
            # self.callback(self, X, attractors, q)
            # move the swarm
            Xprime[..., 0] = X - offset # into generalised coordinates 
            Xprime = (Adt@Xprime[..., None] +  (q[..., None, None]*Sig_L) @
                      np.random.standard_normal(size=(N, D, 2, 1)))[..., 0]
            # apply bounds in physical coordinates
            X = np.clip(Xprime[..., 0] + offset, *bounds.T)
            # recompute attractors
            attractors, scores, offset = self.update_attractors(
                X, scores, attractors[0])
            # callbacks
            self.curve_callback(np.min(scores), np.mean(scores))

        # output pop
        return [zoo.animal(dna=a, fitness=b) for a, b in zip(attractors[0], scores)]


class KrillHerd(continuous_space_optimiser):
    '''
    Krill Herd Optimisation
    Krill move based on three things:
        1) Induced motion
        2) Foraging
        3) Random Physical Diffusion
    dX/dt = N + F + D
    Gandomi, Amir Hossein, and Amir Hossein Alavi. "Krill herd: a new bio-inspired optimization algorithm." Communications in nonlinear science and numerical simulation 17.12 (2012): 4831-4845.
    '''
    name = 'Krill Herd'
    tags = ['Continuous domain', 'Animal', 'Krill Herd']
    hyper_definitions = {
        'N': 'Population size (int)',
        'G': 'Number of generations (int)',
        'Ct': 'Control time step element in (0,2] (float64)',
        'Imotion': 'Inertia coefficients for induced motion (np.array, I.shape=(2,))',
        'Iforage': 'Inertia coefficients for foraging (np.array, I.shape=(2,))',
        'eps': 'epsilon to stop div by 0 errors (small constant) (float64)',
        'Nmax': 'Maximum induced speed in the paper somewhat confusingly (float64)',
        'Vf': 'Foraging speed (float64)',
        'Dmax': 'Maximum diffusion speed in [0.002,0.010] (float64)',
        'Crossover': 'Implement crossover (None or str or Crossover)',
        'Mutate': 'Implement mutation (bool)',
        'Mu': 'Mutation mixing parameter in (0,1) (float64)'
    }
    hyper_defaults = {
        'N': 150,
        'G': 300,
        # NOTE: in the paper this is chosen as a random number in (0,2]
        'Ct': 0.5,
        'Imotion': np.array([0.9, 0.1]),
        'Iforage': np.array([0.9, 0.1]),
        'eps': 1e-12,
        'Nmax': 0.01,
        'Vf': 0.02,
        # NOTE: in the paper this is chosen as a random number in [0.002,0.010]
        'Dmax': 0.005,
        'Crossover': 'binary_crossover',
        'Mutate': True,
        'Mu': 0.5
    }

    def init_pop(self, N):  # TODO make hotswappable with custom bounding
        # Function which initialises the krill randomly within the bounds
        pop = np.empty((N,), dtype=object)
        for i in range(N):
            pop[i] = zoo.krill(
                pos=np.array([np.random.uniform(a, b)
                              for a, b in self.bounds]),
                motion=0.01*np.random.rand(1, self.bounds.shape[0]),
                forage=0.008*np.random.rand(1, self.bounds.shape[0]) + 0.002)
        return pop

    def get_herd(self, pop):
        '''
        It is more convenient to work with matrix representations of the krill
        '''
        D = len(self.bounds)
        vals = np.zeros(self.hypers['N'])
        locs = np.zeros((self.hypers['N'], D))
        motion = np.zeros((self.hypers['N'], D))
        forage = np.zeros((self.hypers['N'], D))
        for i, krill in enumerate(pop):
            vals[i] = krill.fitness
            locs[i, :] = krill.pos
            motion[i, :] = krill.motion
            forage[i, :] = krill.forage
        return [vals, locs, motion, forage]

    def winners_and_losers(self, herd):
        '''
        Sometimes in life you're the best sometimes you're the worst, sorry krill
        '''
        win_idx = np.argmin(herd[0])
        lose_idx = np.argmax(herd[0])
        # Winner and loser are tuples of best/worst (fitness,location)
        winner = (herd[0][win_idx], herd[1][win_idx, :])
        loser = (herd[0][lose_idx], herd[1][lose_idx, :])

        return (winner, loser)

    def all_time_champion(self, pop):
        '''
        Best krill ever
        '''
        best = pop[0].best
        best_pos = pop[0].best_pos
        for p in pop[1:]:
            if p.best < best:
                best = p.best
                best_pos = p.best_pos

        return (best, best_pos)

    def local_motion(self, herd, gen):
        # pairwise distances between krill
        dists = tech.pdist(herd[1])
        # Who's my neighbour
        sense_dist = np.sum(dists, axis=1)/5/self.hypers['N']
        neighbours = dists <= sense_dist
        winner, loser = self.winners_and_losers(herd)
        spread = loser[0] - winner[0]
        if spread == 0:
            warnings.warn(util.KrillSingularityWarning())
            spread = 1e-8
        # Alpha stores local [0] and target [1] for each krill
        alpha = [np.zeros_like(herd[1]), np.zeros_like(herd[1])]
        # Alpha local, the effect of the neighbours
        for i in range(dists.shape[0]):
            Khat = (herd[0][i] - herd[0][neighbours[i, :]]) / spread
            Xhat = (herd[1][neighbours[i, :], :] - herd[1][i, :]) / \
                (dists[i, neighbours[i, :]] + self.hypers['eps'])[:, None]
            alpha[0][i, :] = np.sum(Xhat * Khat[:, None], axis=0)

        # Exploration/exploitation coefficient
        Cbest = 2*(np.random.rand() + gen / self.hypers['G'])

        # Alpha target, take me to your leader
        Kbest = (herd[0] - winner[0]) / spread
        Xbest = (winner[1] - herd[1]) / \
            (dists[:, np.argmin(herd[0])][:, None] + self.hypers['eps'])
        alpha[1] = Cbest*Kbest[:, None]*Xbest

        # Alpha is weighted combination of local and target
        alpha = alpha[0] + alpha[1]

        inertia = tech.lin_reduce(
            self.hypers['Imotion'], gen, self.hypers['G'])
        return self.hypers['Nmax']*alpha + inertia*herd[2], Kbest

    def foraging(self, herd, gen, pop):
        winner, loser = self.winners_and_losers(herd)
        spread = loser[0] - winner[0]
        if spread == 0:
            spread = 1e-8
        # Tasty food at the centre of mass but how good is it
        #Xfood = (np.sum(herd[1]/herd[0][:,None], axis=0) / np.sum(1/herd[0]))[:,None].T
        Xfood = np.average(herd[1], weights=1/herd[0], axis=0)
        Kfood = self.obj(Xfood)
        Xhat_food = (
            Xfood - herd[1]) / (tech.pdist(herd[1], Xfood[None, :]) + self.hypers['eps'])
        Khat_food = (herd[0] - Kfood) / spread
        # Exploration/exploitation coefficient
        Cfood = 2*(1 - gen / self.hypers['G'])
        # Beta food
        beta_food = Cfood * Xhat_food * Khat_food[:, None]
        # Get best previous locations from population
        herd_best = [np.zeros_like(herd[0]), np.zeros_like(herd[1])]
        for i, krill in enumerate(pop):
            herd_best[0][i] = krill.best
            herd_best[1][i, :] = krill.best_pos
        # Xhat and Khat against best previous positions
        Xhat_best = (
            Xfood - herd_best[1]) / (tech.pdist(herd_best[1], Xfood[None, :]) + self.hypers['eps'])
        Khat_best = (herd_best[0] - Kfood) / spread
        # Beta best
        beta_best = Khat_best[:, None]*Xhat_best
        beta = beta_food + beta_best
        # Foraging motion
        inertia = tech.lin_reduce(
            self.hypers['Iforage'], gen, self.hypers['G'])
        return self.hypers['Vf']*beta + inertia*herd[3]

    def random_diffusion(self, gen):
        delta = 2*np.random.rand(self.hypers['N'], len(self.bounds)) - 1
        return self.hypers['Dmax']*(1 - gen/self.hypers['G'])*delta

    def run(self):
        if self.hypers['Crossover'] is not None:
            breeder = self.parse_hyper(self.hypers['Crossover'])
        pop = self.init_pop(self.hypers['N'])
        # Compute first set of fitness
        pop = tech.compute_obj(pop, self.obj)
        # Determine time step as in paper
        if self.bounds is None:
            dt = 10  # If no bounds set use default
        else:
            bounds = self.bounds
            dt = self.hypers['Ct']*np.sum(bounds[:, 1]-bounds[:, 0])
        # Main loop
        for gen in range(self.hypers['G']):
            herd = self.get_herd(pop)
            # Induced motion, following the crowd
            N, Khat_best = self.local_motion(herd, gen)
            # Foraging motion
            F = self.foraging(herd, gen, pop)
            # Random diffusion
            D = self.random_diffusion(gen)
            # Total Velocity
            V = N + F + D
            # Genetic operations
            champion = self.all_time_champion(pop)
            # Crossover
            if self.hypers['Crossover'] is not None:
                crossover_prob = 0.2*Khat_best
                xover_herd = herd[1][np.random.randint(
                    self.hypers['N'], size=(self.hypers['N'],)), :]
                for i, h in enumerate(herd[1]):
                    # Implement 1-to-gbest X-over not 1-to-rand as in paper...
                    xover_herd[i, :] = breeder(
                        h, champion[1], crossover_prob[i])
                current_herd = xover_herd
            else:
                current_herd = herd[1]
            # Mutation
            if self.hypers['Mutate']:
                mutate_prob = Khat_best
                mutate_prob[np.where(Khat_best != 0)] = 0.05 / \
                    Khat_best[np.where(Khat_best != 0)]
                inds = np.random.randint(
                    self.hypers['N'], size=(self.hypers['N'], 2))
                mutates = np.random.rand(self.hypers['N'], len(
                    self.bounds)) < mutate_prob[:, None]
                mutant_dna = champion[1] + self.hypers['Mu'] * \
                    (herd[1][inds[:, 0], :]-herd[1][inds[:, 1], :])
                current_herd[mutates] = mutant_dna[mutates]
            # Move the herd
            new_pos = current_herd + dt*V
            # Compute objectives and update the herd
            for i, (dna, motion, forage) in enumerate(zip(new_pos, N, F)):
                pop[i].pos = dna
                pop[i].motion = motion
                pop[i].forage = forage
            self.apply_bounds(pop)
            tech.compute_obj(pop, self.obj)
        return pop



'''
Standard / common techniques that are used in several optimisers are abstracted to functions here. 
'''

# %% Imports

import numpy as np

from freelunch.zoo import animal


# %% Common methods




def compute_obj(pop, obj):
    for sol in pop:
        sol.fitness = obj(sol.dna)
    return pop


def bounds_as_mat(bounds):
    bounds_mat = np.zeros((len(bounds), 2))
    for i, bound in enumerate(bounds):
        bounds_mat[i, 0], bounds_mat[i, 1] = bound
    return bounds_mat



def lin_reduce(lims, n, n_max):
    # Linearly reduce with generations, e.g. inertia values
    if lims[1] < lims[0]: 
        if isinstance(lims,list):
            lims.reverse()
        else:
            np.flip(lims)
    return lims[1] + (lims[0]-lims[1])*n/n_max


def pdist(A, B=None):
    '''
    Pairwise Euclidean Distance inside array
    '''
    if B is None:
        B = A
    return np.sqrt(np.sum((A[:, None]-B[None, :])**2, axis=-1))


#%% Bounding Strategies


def no_bounding(p, bounds, **hypers):
    '''
    Placeholder for no bounding
    '''
    raise Warning('No bounds applied')

def sticky_bounds(p, bounds, eps=1e-12, **hypers):
    '''
    Apply sticky bounds to space
    '''
    out = p.dna[:]
    for i, bound in enumerate(bounds):
        low, high = bound
        if   p.dna[i] > high: out[i] = high - eps
        elif p.dna[i] < low:  out[i] = low + eps
        p.dna = out

# %% Intialisation strategies

def uniform_continuous_init(bounds, N, creature=animal):
    out = np.empty((N,), dtype=object)
    for i in range(N):
        adam = creature()
        adam.dna = np.array([np.random.uniform(a, b)
                             for a, b in bounds])
        out[i] = adam
    return out

def Gaussian_neigbourhood_init(bounds, N, creature=animal, mu=None, sig=None):
    if mu is None:
        mu = [(a+b)/2 for a,b in bounds]
    if sig is None:
        sig = [(b-a)/6 for a,b in bounds]
    out = np.empty((N,), dtype=object)
    for i in range(N):
        adam = creature()
        adam.dna = np.random.normal(mu, sig)
        out[i] = adam
    return out


'''
Base classes for optimisers

'''
from functools import partial
from multiprocessing import Pool

import numpy as np

from freelunch import darwin
from freelunch import tech
from freelunch.adaptable import adaptable_set
from freelunch.util import UnpicklableObjectiveFunction, _tolist


class optimiser:
    '''
    Base class for all optimisers
    '''
    name = 'optimiser'
    tags = []
    hyper_definitions = {}
    hyper_defaults = {}
    can_run_quick = False

    def __init__(self, obj, hypers={}, bounds=None):
        '''
        Unlikely to instance the base class but idk what else goes here
        '''
        self.bounds = bounds  # Bounds / constraints
        self.bounder = tech.sticky_bounds
        self.initialiser = tech.uniform_continuous_init 
        self.nfe = 0
        # Extend capability of objective function (mp safe!!)
        self.obj = partial(self._obj, obj)
        # Hyperparamters/ methods
        self.hypers = dict(self.hyper_defaults, **hypers)
        self.hypers['bounding'] = self.bounder.__name__
        self.hypers['initialisation'] = self.initialiser.__name__

    def __call__(self, n_runs=1, n_return=1, full_output=False, n_workers=1, pool_args={}, chunks=1):
        '''
        API for running the optimisation
        '''
        if n_runs > 1:
            if n_workers > 1:
                try:
                    ret = Pool(n_workers, **pool_args).starmap(self.run_mp,
                                                               [() for _ in range(n_runs)], chunks)
                except AttributeError:
                    raise UnpicklableObjectiveFunction
                runs, _nfes = [list(a) for a in zip(*ret)]
                self.nfe = sum(_nfes)
            else:
                runs = [self.run() for i in range(n_runs)]
            sols = np.concatenate(runs)
        else:
            sols = self.run()
        sols = sorted(sols)
        if not full_output:
            return np.array([sol.dna for sol in sols[:n_return]])
        else:
            return {
                'optimiser': self.name,
                'hypers': {k: _tolist(v) for k, v in self.hypers.items()},
                'bounds': _tolist(self.bounds),
                'nruns': n_runs,
                'nfe': self.nfe,
                'solutions': [_tolist(sol.dna) for sol in sols],
                'scores': [sol.fitness for sol in sols]
            }

    def __repr__(self):
        return self.name + ' optimisation object'

    def run_mp(self):
        return self.run(), self.nfe

    def parse_hyper(self, op):
        if isinstance(op, list):  # top 10 recursive gamer moments
            strats = [self.parse_hyper(strat) for strat in op]
            return adaptable_set(strats)
        elif isinstance(op, type) and issubclass(op, darwin.genetic_operation):
            return op()
        elif isinstance(op, str):
            try:
                op = getattr(darwin, op)
                return op()
            except AttributeError:
                raise AttributeError(
                    'Method not recognised, refer to docs for list of implemented methods')  # TODO test

    def apply_bounds(self, pop, **hypers):
        '''Apply the bouding method to every object in an iterable'''
        for sol in pop:
            self.bounder(sol, self.bounds, **hypers)

    def _obj(self, obj, vec):
        '''Adds nfe counting and bad value handling to raw_obj'''
        self.nfe += 1
        fit = obj(vec)
        try:
            if np.isnan(fit):
                return None
            return float(fit)
        except(ValueError, TypeError):
            return None

    def curve_callback(self, best=None, mean=None):
        if best is None:
            self.data = {'best':[], 'mean':[]}
        else:
            self.data['best'].append(best)
            self.data['mean'].append(mean)

# Subclasses for granularity

class continuous_space_optimiser(optimiser):
    '''
    Base class for continuous space optimisers i.e DE, PSO, SA etc.
    '''


class discrete_space_optimiser(optimiser):
    '''
    Base class for discrete space optimisers i.e GA, GP etc.
    '''



'''
Adaptable parameters and methods, functionality and API
'''

import numpy as np


# %% Adaptable methods


class adaptable_method():
    '''
    Base class for methods / parameters that are selected / updated as part of an adaptive scheme
    '''
    name = 'adaptable item'

    def __init__(self):
        self.hits = [0]
        self.wins = [0]
        self.p = [1]  # probability of selection

    def __call__(self, *parents, **hypers):
        self.hits[-1] += 1
        return self.op(*parents, **hypers)

    def win(self):
        self.wins[-1] += 1

    def reset_counts(self):
        self.hits.append(0)
        self.wins.append(0)

    def op(self,*parents, **hypers):
        raise NotImplementedError

class adaptable_set():
    '''
    Class for sets of competing methods i.e the mutation ops in SADE.

    Contains api for pinwheel selection and probability update
    '''
    name = 'adaptable set of methods'

    def __init__(self, strats=[]):
        self.strats = strats

    def select_strategy(self):
        '''
        API for pinwheel selection
        '''
        ps = [s.p[-1] for s in self.strats]
        ps = ps/np.sum(ps)
        return np.random.choice(self.strats, p=ps)

    def update_strategy_ps(self):
        '''
        API for probability update
        '''
        hits = np.array([s.hits[-1] for s in self.strats])
        wins = np.array([s.wins[-1] for s in self.strats])
        total_hits = np.sum(hits)
        total_wins = np.sum(wins)
        ps = np.zeros_like(hits)
        # update model - maybe abstract this at some point
        # prevent zero division with a bit of fun bias!!
        dem = np.sum(wins * (hits + wins)) + 1
        for i, h, w in zip(range(len(ps)), hits, wins):
            num = h * (total_hits + total_wins)
            ps[i] = num / dem
        # normalise
        n = np.sum(ps)
        if n == 0:
            ps += 1
        ps = ps / n
        for strat, p in zip(self.strats, ps):
            strat.reset_counts()
            strat.p.append(p)


# %% Adaptable parameter classes and derivatives


class adaptable_parameter():
    '''
    Boiler plate for parameter
    '''

    def __init__(self, value=None):
        self.value = value
        self.hits = [0]
        self.wins = [0]
        self.win_values = []

    def __call__(self, *args):
        self.hits[-1] += 1
        return self.op(*args)

    def now(self):
        # return tuple of self and current value
        return self, self.value

    def op(self):
        return self.value

    def win(self, v):
        '''adaptable parameters store successful values'''
        self.win_values.append(v)
        self.wins[-1] += 1


class linearly_varying_parameter(adaptable_parameter):

    def __init__(self, a0, an, n):
        super().__init__()
        self.a0 = a0
        self.an = an
        self.n = n
        self.values = np.linspace(a0, an, n)

    def op(self, k):
        return self.values[k]


class normally_varying_parameter(adaptable_parameter):

    def __init__(self, u, sig):
        super().__init__()
        self.u = u
        self.sig = sig
        self.value = np.random.normal(self.u, self.sig)

    def op(self):
        self.value = np.random.normal(self.u, self.sig)
        return self.value

    def update(self):  # fit normal distribution to successful parameters
        u = self.u if len(self.win_values) == 0 else np.mean(self.win_values)
        std = self.std if len(self.win_values) == 0 else max(np.std(self.win_values), 10**-3)
        if not np.isnan(u):
            self.u = u 
        else:
            raise ValueError # Stop don't just stay still
        if not np.isnan(std):
            self.std = std
        else:
            raise ValueError # Stop don't just stay still
        self.win_values = [] # Not storing all winning parameters at the moment
        self.wins.append(0)
        self.hits.append(0)



'''
New home for the zoo of optimisation creatures
'''
import numpy as np

from freelunch.util import InvalidSolutionUpdate, BadObjectiveFunctionScores
from freelunch.util import verify_real_finite


# %% Base class

class animal:
    '''
    Handy dandy common object for storing trial solutions / other interesting data
    '''

    def __init__(self, dna=None, fitness=None, best=None, best_pos=None):
        
        
        self.best= best
        self.best_pos = best_pos
        self.dna = dna
        self._fitness = None
        self.fitness = fitness
        self.tech = []

    def move(self, dna, fitness):
        if np.all(np.isreal(dna)) and\
             not np.any(np.isinf(dna)) and\
             not np.any(np.isnan(dna)):
            self.dna = dna
            self.fitness = fitness
        else:
            raise InvalidSolutionUpdate

    @property
    def fitness(self):
        return self._fitness

    
    @fitness.setter
    @verify_real_finite([1], ['fitness'])
    def fitness(self, fitness):
        if (fitness is not None) and ((self._fitness is None) or (fitness < self.best)):
            self.best = fitness
            self.best_pos = self.dna
        self._fitness = fitness
    
    def on_win(self):
        '''big brain time'''
        for t in self.tech:
            if isinstance(t, tuple):
                t, v, = t
                t.win(v)
            else:
                t.win()
        self.tech = []

    def __lt__(self, other):
        '''overlaod the < operator for convenient handling of tournament selction in the presence onf nonetype fitnesses'''
        if self.fitness is None:
            if other.fitness is None:
                # ? what to do here - raise error TR
                # return False
                raise BadObjectiveFunctionScores
            return False # Other has lower fitness
        elif other.fitness is None:
            return True # We have the lower fitness            
        else:
            return self.fitness < other.fitness
        
    def __gt__(self, other):
        '''overlaod the > operator for convenient handling of tournament selction in the presence onf nonetype fitnesses'''
        if self.fitness is None:
            if other.fitness is None:
                # ? what to do here - raise error TR
                # return False
                raise BadObjectiveFunctionScores
            return True # Other has lower fitness
        elif other.fitness is None:
            return False # We have the lower fitness            
        else:
            return self.fitness > other.fitness

    def __eq__(self, o: object) -> bool:
        # Not the same class
        if type(self) != type(o):
            return False
        # Get user attributes from self and o
        us = {k: v for k, v in self.__dict__.items()}
        them = {k: v for k, v in o.__dict__.items()}

        for k in us.keys():
            if np.any(us[k] != them[k]):
                return False
        
        # If not false must be true
        return True
    
    def as_sol(self):
        return animal(dna=self.best_pos, fitness=self.best)


# %% All that the light touches is our domain
class particle(animal):
    '''
    Want to store info on particles in a swarm? I got you bud
    '''

    def __init__(self, pos=None, vel=None, fitness=None, best=None, best_pos=None):
        
        super().__init__(dna=pos, fitness=fitness, best=best, best_pos=best_pos)
        self.vel = vel

    
    def move(self, pos, vel, fitness):
        if np.all(np.isreal(pos)) and\
             not np.any(np.isinf(pos)) and\
             not np.any(np.isnan(pos)) and\
             np.all(np.isreal(vel)) and\
             not np.any(np.isinf(vel)) and\
             not np.any(np.isnan(vel)):
            self.pos = pos
            self.vel = vel
            self.fitness = fitness
        else:
            raise InvalidSolutionUpdate

    @property
    def dna(self):
        return self.pos
    
    @dna.setter
    def dna(self, dna):
        self.pos = dna       


class krill(particle):

    '''
    I am a krill, a type of animal 

    I am also basically just a particle...
    '''

    def __init__(self, pos=None, vel=None, fitness=None, best=None, best_pos=None, motion=None, forage=None):
        
        super().__init__(pos=pos, vel=vel, fitness=fitness, best=best, best_pos=best_pos)

        self.motion = motion
        self.forage = forage



from setuptools import setup


# The text of the README file
with open('README.md') as f:
    rm = f.read()

# This call to setup() does all the work
setup(
    name="freelunch",
    version="0.0.15",
    description="Heuristic and meta-heuristic optimisation suite in Python",
    long_description=rm,
    long_description_content_type="text/markdown",
    url="https://github.com/MDCHAMP/FreeLunch",
    author="Max Champneys",
    author_email="max.champneys@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
    ],
    packages=['freelunch'],
    package_dir={'':'src'},
    include_package_data=False,
    install_requires=[
        "numpy"
    ],
)


'''
Utility functions
'''

import numpy as np

# %% Custom exceptions

class BadObjectiveFunctionScores(Exception):
    '''Exception raised when both objective score comparisons evaluate false'''

class UnpicklableObjectiveFunction(Exception):
    '''Exception raised when objectivefunction cannot be pickled for multiprocessing'''
class ZeroLengthSolutionError(Exception):
    '''Exception raised when an empty solution is passed to a benchmark'''
class InvalidSolutionUpdate(Exception):
    '''Exception raised when trying to move an animal/particle to a bad location'''

# %% Custom warnings

class SolutionCollapseWarning(Warning):
    '''Warning: All solutions in the population are identical'''

class KrillSingularityWarning(SolutionCollapseWarning):
    '''Warning: Krill singularity identified, adding pertubation'''

#%% Helper functions

def real_finite(a):

    if a is None:
        return
    elif not isinstance(a, float) and not isinstance(a,int):
        raise ValueError
    elif not np.isreal(a):
        raise ValueError
    elif np.isnan(a):
        raise ValueError
    elif a == np.inf or a == -np.inf:
        raise ValueError


def expm(A):
    v,S = np.linalg.eig(A)
    if not len(np.unique(v))==len(v):
        raise ValueError('Non-diagonisable input matrix! Try choosing different parameters')
    return np.real(S@np.diag(np.exp(v))@np.linalg.inv(S))

def _tolist(arr):
    if isinstance(arr, np.ndarray): return arr.tolist()
    else: return arr


#%% Decorators

def verify_real_finite(check_args, check_vars):

    def wrapped_func(func):
        def wrapper(*args, **kwargs):
            for a in check_args:
                real_finite(args[a])
            for v in check_vars:
                try:
                    real_finite(kwargs[v])
                except KeyError:
                    pass
            func(*args, **kwargs)
        return wrapper
    return wrapped_func